#if ! defined(__LKT_DATABASE_H__)
#define __LKT_DATABASE_H__

#include <lektor/common.h>
#include <lektor/mkv.h>
#include <lektor/uri.h>

#include <sqlite3.h>
#include <mpv/client.h>

#include <stdbool.h>
#include <linux/limits.h>

struct lkt_queue_state {
    int volume;
    int paused;
    int random;
    int repeat;
    int single;
    int current;
    int duration;
    int consume;
    int length;
};

/* Update stuff */
void database_get_update  (volatile sqlite3 *db, long *timestamp, long *job, int *current);
void database_stamp       (volatile sqlite3 *db);
void database_updated     (volatile sqlite3 *db);
void database_deleted_kara(volatile sqlite3 *db, int **kara_id, size_t *len);

/* Open correctly a database for lektor. */
bool database_new   (volatile sqlite3 **db);
bool database_init  (const char *dbpath);
bool database_open  (volatile sqlite3 *db, const char *dbpath, bool check);
bool database_attach(volatile sqlite3 *db, const char *name, const char *dbpath);
bool database_detach(volatile sqlite3 *db, const char *name);
void database_close_all(void);

/* Get information on the queue and currently playing kara. */
bool database_queue_state       (volatile sqlite3 *db, struct lkt_queue_state *res);
bool database_queue_current_kara(volatile sqlite3 *db, struct kara_metadata *res, int *id);

bool database_queue_set_paused       (volatile sqlite3 *db, bool paused);
bool database_queue_set_current_index(volatile sqlite3 *db, int idx);
bool database_queue_get_current_file (volatile sqlite3 *db, char filepath[PATH_MAX]);

bool database_get_kara_path(volatile sqlite3 *db, int id, char filepath[PATH_MAX]);

/* Update the database. */
bool database_update              (volatile sqlite3 *db, const char *kara_dir, int check_timestamp);
bool database_update_add          (volatile sqlite3 *db, const char *kara_path,
                                   struct kara_metadata *mdt, uint64_t id, bool avail);
bool database_update_set_available(volatile sqlite3 *db, int id);
void database_update_del          (volatile sqlite3 *db, int id);
void database_update_touch        (volatile sqlite3 *db, int id);

/* Control the content of the queue. */
bool database_queue_add_uri(volatile sqlite3 *db, struct lkt_uri *uri, int priority);
bool database_queue_add_id (volatile sqlite3 *db, int id,              int priority);
bool database_queue_del_id (volatile sqlite3 *db, int id);
bool database_queue_del_pos(volatile sqlite3 *db, int pos);
bool database_queue_clear  (volatile sqlite3 *db);
bool database_queue_crop   (volatile sqlite3 *db);
bool database_queue_move   (volatile sqlite3 *db, int from, int to);
bool database_queue_shuffle(volatile sqlite3 *db);
bool database_queue_seekid (volatile sqlite3 *db, int id, int *out_pos);
bool database_queue_dump   (volatile sqlite3 *db, const char *plt_name);

/* Control the playing state of the queue. */
bool database_queue_toggle_pause(volatile sqlite3 *db);
bool database_queue_play        (volatile sqlite3 *db, int pos);
bool database_queue_stop        (volatile sqlite3 *db);

/* A search callback to be called after each matched row */
struct lkt_callback {
    bool (*call)(void *args, int pos, int pos_len, int id, int id_len, const char *sql_row);
    struct lkt_state *srv;
    size_t c;
    int iterations;
};

/* List the content of the queue */
bool database_queue_list(volatile sqlite3 *db, size_t from, size_t to,
                         struct lkt_callback *callback);

/* Search the database */
struct lkt_search {
    volatile sqlite3 *db;
    sqlite3_stmt *stmt;
    enum lkt_search_type {
        lkt_search_database,
        lkt_search_playlist,
        lkt_search_queue,
        lkt_search_sticker,
        lkt_search_listplaylist,
    } type;

    void (*call)(void);     /* Called during the iter phase, casted     */

    struct lkt_state *srv;
    size_t c;

    long continuation;      /* The continuation state of the client     */
    int msg_count;          /* How much messages we can send            */

    union {
        const char *st_name;    /* Sticker name                         */
        const char *plt_name;   /* Playlist name                        */
    };
    int st_value;               /* The value of a sticker               */
    int st_uri;                 /* URI of a sticker                     */
    char st_op;                 /* Comparaison operator for stickers    */
    char *st_type;              /* Type of sticker                      */

    struct lkt_uri ka_uri;      /* Search by uri                        */
};

typedef bool (*lkt_search_init_add_func)    (volatile sqlite3 *, struct lkt_uri *, int);
typedef bool (*lkt_search_database_func)    (struct lkt_state *srv, size_t c, int id, int id_len,
        const char *row);
typedef bool (*lkt_search_queue_func)       (struct lkt_state *srv, size_t c, int id, int id_len,
        const char *row);
typedef bool (*lkt_search_sticker_func)     (struct lkt_state *srv, size_t c, const char *sticker,
        const char *type, int uri, int value);
typedef bool (*lkt_search_listplaylist_func)(struct lkt_state *srv, size_t c, const char *name);

bool database_search_database_init    (volatile sqlite3 *db, struct lkt_search *ret);
bool database_search_queue_init       (volatile sqlite3 *db, struct lkt_search *ret);
bool database_search_sticker_init     (volatile sqlite3 *db, struct lkt_search *ret);
bool database_search_playlist_init    (volatile sqlite3 *db, struct lkt_search *ret);
bool database_search_listplaylist_init(volatile sqlite3 *db, struct lkt_search *ret);
bool database_search_iter(struct lkt_search *item);
bool database_kara_by_id (volatile sqlite3 *db, int id, struct kara_metadata *kara,
                          char filepath[PATH_MAX]);

/* Next and prev operation on the queue. */
bool database_queue_next(volatile sqlite3 *db, char filepath[PATH_MAX]);
bool database_queue_prev(volatile sqlite3 *db, char filepath[PATH_MAX]);
bool database_queue_skip_current(volatile sqlite3 *db, char filepath[PATH_MAX]);

/* Set a value in the config table */
bool database_validate_conf  (volatile sqlite3 *db);
bool database_config_set     (volatile sqlite3 *db, const char *section, const char *key,
                              const char *value);
bool database_config_get_text(volatile sqlite3 *db, const char *section, const char *key,
                              char *value, size_t len);
bool database_config_get_text_nospace(volatile sqlite3 *db, const char *section,
                                      const char *key, char *value, size_t len);
bool database_config_get_int (volatile sqlite3 *db, const char *section, const char *key,
                              int *value);
bool database_config_exists  (volatile sqlite3 *db, const char *section, const char *key);
bool database_config_queue   (volatile sqlite3 *db, const char *option,  int value);
bool database_config_queue_default(volatile sqlite3 *db);

/* Get a value from the config table */
bool database_get_config(volatile sqlite3 *db, const char *option, int *value);

/* Playlists */
struct lkt_playlist_metadata {
    const char *name;
    int last_update;
    int content_count;
};

/* Control playlists */
bool database_plt_create    (volatile sqlite3 *db, const char *name);
bool database_plt_remove    (volatile sqlite3 *db, const char *name);
bool database_plt_remove_pos(volatile sqlite3 *db, const char *name, int pos);
bool database_plt_clear     (volatile sqlite3 *db, const char *name);
bool database_plt_rename    (volatile sqlite3 *db, const char *old_name, const char *new_name);

bool database_plt_export(volatile sqlite3 *db, const char *name);
bool database_plt_import(volatile sqlite3 *db, const char *name);

bool database_plt_add_uri(volatile sqlite3 *db, const char *name, struct lkt_uri *uri);

/* User control, yeah, MPD authentification sucks. */
bool database_user_authentificate(volatile sqlite3 *db, const char *password);
bool database_user_add(volatile sqlite3 *db, const char *username, const char *password);

/* Stickers manipulations. */
bool database_sticker_create        (volatile sqlite3 *db, const char *name);
bool database_sticker_delete        (volatile sqlite3 *db, const char *name);
bool database_sticker_delete_specify(volatile sqlite3 *sb, const char *type, int uri,
                                     const char *name);
bool database_sticker_set           (volatile sqlite3 *db, const char *type, const char *name,
                                     int uri,
                                     int value);

#endif /* __LKT_DATABASE_H__ */
