#if ! defined(__LKT_QUEUE_H__)
#define __LKT_QUEUE_H__
#define _POSIX_C_SOURCE 200809L

/* Everything in this file is implemented in the header, so that modules don't
   have to compile the lib with another .c file. */

#include <pthread.h>
#include <errno.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <lektor/common.h>

enum {
    __LKT_PLAY_STOP   = 0,
    __LKT_PLAY_PLAY   = 1,
    __LKT_PLAY_PAUSE  = 2,
    __LKT_PLAY_TOGGLE = 3,
};

enum {
    __LKT_DB_UPDATING_FINISHED = 0, /* Not updating       */
    __LKT_DB_UPDATING_PROGRESS = 1, /* Update in progress */
};

#define LKT_PLAY_STOP   ((void *) (size_t) __LKT_PLAY_STOP)
#define LKT_PLAY_PLAY   ((void *) (size_t) __LKT_PLAY_PLAY)
#define LKT_PLAY_PAUSE  ((void *) (size_t) __LKT_PLAY_PAUSE)
#define LKT_PLAY_TOGGLE ((void *) (size_t) __LKT_PLAY_TOGGLE)

#define LKT_DB_UPDATING_PROGRESS ((void *) (size_t) __LKT_DB_UPDATING_PROGRESS)
#define LKT_DB_UPDATING_FINISHED ((void *) (size_t) __LKT_DB_UPDATING_FINISHED)

enum lkt_event_type {
    lkt_event_null          = 0,         // NULL
    lkt_event_play_pos      = (1 <<  1), // size_t
    lkt_event_play_file     = (1 <<  2), // XXX: UNUSED //
    lkt_event_play_next     = (1 <<  3), // NULL
    lkt_event_play_prev     = (1 <<  4), // NULL
    lkt_event_play_toggle   = (1 <<  5), // size_t, `LKT_PLAY_.*`
    lkt_event_prop_vol      = (1 <<  6), // size_t
    lkt_event_prop_dur      = (1 <<  7), // size_t
    lkt_event_prop_time     = (1 <<  8), // size_t
    lkt_event_skip_current  = (1 <<  9), // NULL
    lkt_event_db_updating   = (1 << 10), // size_t, `LKT_DB_UPDATING_*`
};

#define lkt_event_play ( lkt_event_play_pos    | lkt_event_play_file    \
                       | lkt_event_play_next   | lkt_event_play_prev    \
                       | lkt_event_play_toggle | lkt_event_skip_current )
#define lkt_event_prop ( lkt_event_prop_vol    | lkt_event_prop_dur     \
                       | lkt_event_prop_time                            )

typedef struct {
    unsigned int type;
    void *attr;
} lkt_event;

struct queue {
    pthread_mutex_t lock;

    volatile lkt_event *contents;
    volatile size_t size;
    volatile size_t last;

    volatile int available;
};

static inline int
lkt_queue_new(struct queue *ret)
{
    if (!ret)
        return 1;

    pthread_mutex_t mxt = PTHREAD_MUTEX_INITIALIZER;
    struct queue _ret = {
        .contents = malloc(LKT_DEFAULT_LIST_SIZE * sizeof(lkt_event)),
        .size = LKT_DEFAULT_LIST_SIZE,
        .last = 0,
        .available = 0,
        .lock = mxt,
    };

    if (_ret.contents == NULL)
        return 1;

    *ret = _ret;
    return 0;
}

static inline void
lkt_queue_free(struct queue *queue)
{
    pthread_mutex_lock(&queue->lock);
    if (queue && queue->contents)
        free((void *) queue->contents);
    pthread_mutex_unlock(&queue->lock);
}

static inline void
lkt_queue_send(struct queue *queue, enum lkt_event_type _type, void *_attr)
{
    pthread_mutex_lock(&queue->lock);
    GOTO_UNLESS(queue, "Invalid argument", end)

    volatile lkt_event *new;
    if (queue->size == queue->last) {
        new = realloc((void *) queue->contents, queue->size * 2 * sizeof(lkt_event));

        if (NULL == new)
            goto end;

        queue->contents = new;
        queue->size *= 2;
    }

    lkt_event evt = {
        .type = _type,
        .attr = _attr,
    };
    queue->contents[(queue->last)++] = evt;

    if (! (_type & queue->available))
        LOG_WARN("QUEUE", "The event %d is not available, push it anyway", _type);
end:
    pthread_mutex_unlock(&queue->lock);
}

static inline lkt_event
lkt_queue_handle(struct queue *queue)
{
    pthread_mutex_lock(&queue->lock);
    lkt_event ret = {0};
    if (!queue || !queue->last)
        goto end;
    if (! (queue->contents[0].type & queue->available) &&
        queue->contents[0].type) {
        LOG_WARN("QUEUE", "Event %d is not available", queue->contents[0].type);
        goto end;
    }

    ret = queue->contents[0];
    memmove((void *) queue->contents, (void *) (queue->contents + 1),
            --(queue->last));
end:
    pthread_mutex_unlock(&queue->lock);
    return ret;
}

static inline void
lkt_queue_make_available(struct queue *queue, enum lkt_event_type type)
{
    pthread_mutex_lock(&queue->lock);
    queue->available |= type;
    pthread_mutex_unlock(&queue->lock);
}

#endif /* __LKT_QUEUE_H__ */
