#define _POSIX_C_SOURCE 200809L

#include <lektor/common.h>
#include <lektor/database.h>

#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <sqlite3.h>

#include "macro.h"

bool
database_config_set(volatile sqlite3 *db, const char *section, const char *key,
                    const char *value)
{
    static const char *SQL_STMT =
        "INSERT OR REPLACE INTO"
        " config (section, key, value)"
        " VALUES (?, ?, ?);\n";
    sqlite3_stmt *stmt = 0;
    bool ret = false;
    int code;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);
    SQLITE_BIND_TEXT(db, stmt, 3, value, error);

    code = sqlite3_step(stmt);

    if (code != SQLITE_OK && code != SQLITE_DONE) {
        LOG_ERROR("DB", "Failed to insert or replace: %s",
                  sqlite3_errmsg((sqlite3 *) db));
        goto error;
    }

    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_get_text(volatile sqlite3 *db, const char *section,
                         const char *key, char *value, size_t len)
{
    static const char *SQL_STMT =
        "SELECT value"
        " FROM config"
        " WHERE section = ? AND key = ?;\n";
    sqlite3_stmt *stmt = 0;
    bool ret = false;
    char *row;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);
    SQLITE_STEP_ROW(db, stmt, error);

    row = (char *) sqlite3_column_text(stmt, 0);
    strncpy(value, row, len);
    value[len - 1] = 0;
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_get_text_nospace(volatile sqlite3 *db, const char *section, const char *key,
                                 char *value, size_t len)
{
    static const char *SQL_STMT =
        "SELECT REPLACE(value, ' ', '')"
        " FROM config"
        " WHERE section = ? AND key = ?;\n";
    sqlite3_stmt *stmt = 0;
    bool ret = false;
    char *row;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);
    SQLITE_STEP_ROW(db, stmt, error);

    row = (char *) sqlite3_column_text(stmt, 0);
    strncpy(value, row, len);
    value[len - 1] = 0;
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_exists(volatile sqlite3 *db, const char *section,
                       const char *key)
{
    static const char *SQL_STMT =
        "SELECT value"
        " FROM config"
        " WHERE section = ? AND key = ?;\n";
    sqlite3_stmt *stmt = 0;
    bool ret = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);
    SQLITE_STEP_ROW(db, stmt, error);

    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_get_int(volatile sqlite3 *db, const char *section,
                        const char *key, int *value)
{
    static const char *SQL_STMT =
        "SELECT CAST(value AS INTEGER)"
        " FROM config"
        " WHERE section = ? AND key = ?;\n";
    sqlite3_stmt *stmt = 0;
    bool ret = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, section, error);
    SQLITE_BIND_TEXT(db, stmt, 2, key, error);

    SQLITE_STEP_ROW(db, stmt, error);

    *value = sqlite3_column_int(stmt, 1);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_config_queue(volatile sqlite3 *db, const char *option, int value)
{
    static const char *SQL_STMT_TMP =
        "UPDATE queue_state SET %s = ? WHERE id = 42;";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];
    sqlite3_stmt *stmt = 0;
    bool ret = false;
    int code;

    if (value < 0)
        value = 0;
    if (value > 100)
        value = 100;

    safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL_STMT_TMP, option);
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_INT(db, stmt, 1, value, error);

    code = sqlite3_step(stmt);

    if (code != SQLITE_OK || code != SQLITE_ROW || code != SQLITE_DONE)
        goto error;

    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_get_config(volatile sqlite3 *db, const char *option, int *value)
{
    static const char *SQL_STMT_TMP =
        "SELECT %s FROM queue_state WHERE id = 42;";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];
    sqlite3_stmt *stmt = 0;
    bool ret = false;

    safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL_STMT_TMP, option);
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_STEP_ROW(db, stmt, error);
    *value = sqlite3_column_int(stmt, 0);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_validate_conf(volatile sqlite3 *db)
{
    const char *section;
#define section(_sct)   section = _sct;
#define value_opt(name, value)
#define value(name, value)                                                    \
    if (!database_config_exists(db, section, name)) {                         \
        LOG_ERROR("CONFIG", "Missing option \""name"\" in section \"%s\"",\
                      section);                                               \
        return false;                                                         \
    }
#include <lektor/config.def>
#undef section
#undef value
#undef value_opt
    return true;
}
