#define _POSIX_C_SOURCE 200809L

#include <lektor/lktconfig.h>
#include <lektor/common.h>
#include <lektor/config.h>
#include <lektor/net.h>
#include <lektor/cmd.h>
#include <lektor/reg.h>
#include <lektor/database.h>
#include <lektor/commands.h>

#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <pthread.h>
#include <sys/stat.h>

#if defined (LKT_STATIC_MODULE)
REG_DECLARE(sdl2_reg)
REG_DECLARE(repo_reg)
#endif

/* Recursive mkdir, where the last word of the string is a file, not a folder. */
static inline void
__mkdir(const char *dir)
{
    char tmp[PATH_MAX];
    char *p = NULL;
    safe_snprintf(tmp, sizeof(tmp), "%s", dir);
    size_t len = strlen(tmp);
    /* In our case, the final word is always a file, not a folder. */
    if (tmp[len - 1] == '/')
        tmp[len - 1] = 0;
    for (p = tmp + 1; *p; p++) {
        if(*p == '/') {
            *p = 0;
            mkdir(tmp, 00700);
            *p = '/';
        }
    }
    /* Don't do final mkdir here, because in our case the final word in the string
     * is a file, not a folder.
     * mkdir(tmp, S_IRWXU); */
}

int
main(int argc, char *argv[])
{
    REG_BEGIN(server_reg)
#if defined (LKT_STATIC_MODULE)
    REG_REGISTER("repo", repo_reg)
    REG_REGISTER("sdl2", sdl2_reg)
#endif
    REG_END()

    char exe[PATH_MAX];
    int autoclear, check_exclusive = 1, opt;
    char *conf_file = safe_zero_malloc(PATH_MAX * sizeof(char));
    executable_name = "lektord";

    /* Check args */
    while ((opt = getopt(argc, argv, "hFf:")) != -1) {
        switch (opt) {
        case 'F':
            check_exclusive = 0;
            break;
        case 'f':
            strncpy(conf_file, optarg, PATH_MAX);
            break;
        case 'h':
        default:
            print_help();
            exit(EXIT_SUCCESS);
        }
    }

    reg_export(server_reg);
    if (read_self_exe(exe, PATH_MAX))
        LOG_WARN("GENERAL", "Failed to read self executable path, "
                 "restart may not work");
    executable_name = exe;

    /* Init the server */
    char *db_path   = safe_malloc(PATH_MAX * sizeof(char));
    char *kara_dir  = safe_malloc(PATH_MAX * sizeof(char));
    struct lkt_state srv = { .kara_prefix = kara_dir };
    if (lkt_queue_new(&srv.queue)) {
        LOG_ERROR("INIT", "Faield to create server queue");
        exit(EXIT_FAILURE);
    }

    /* Initialize the system. */
    if (!database_new(&srv.db)) {
        LOG_ERROR("INIT", "Failed to init memory db");
        exit(EXIT_FAILURE);
    }

    /* Read or create default config file. */
    int retry_config_once = 0;
retry_config:
    if (conf_file[0] == '\0' && config_detect_file(conf_file, PATH_MAX)) {
        if (retry_config_once) {
            LOG_ERROR("INIT", "Failed to find a config file");
            exit(EXIT_FAILURE);
        } else {
            LOG_INFO("INIT", "Creating default config file");
            config_default_file(conf_file, PATH_MAX);
            __mkdir(conf_file); /* Create the folder for the file. */
            FILE *file_desc = fopen(conf_file, "w+");
            if (NULL == file_desc) {
                LOG_ERROR("FATAL", "Failed to open default config file '%s' to initialize it",
                          conf_file);
                exit(EXIT_FAILURE);
            }
            config_default(file_desc);
            fclose(file_desc);
            LOG_INFO("INIT", "Default configuration file has been writen to %s", conf_file);
            retry_config_once = 1;
            goto retry_config;
        }
    }

    if (config_new(srv.db, conf_file)) {
        LOG_ERROR("INIT", "Failed to read the config");
        exit(EXIT_FAILURE);
    }
    free(conf_file);

    /* Finish to initialize. */
    database_config_get_text(srv.db, "database", "db_path", db_path, PATH_MAX);
    RETURN_UNLESS(database_open(srv.db, db_path, check_exclusive), "Can't open database", 1);
    free(db_path);

    /* Read the configuration. We already know that the config is valid */
    database_config_get_int (srv.db, "player",   "autoclear", &autoclear);
    database_config_get_text(srv.db, "database", "kara_dir",  kara_dir, PATH_MAX);
    database_config_get_text(srv.db, "server",   "host",      srv.host, HOST_NAME_MAX);
    database_config_get_text(srv.db, "server",   "port",      srv.port, 5);

    if (kara_dir[strlen(kara_dir) - 1] != '/')
        strncat(kara_dir, "/", PATH_MAX - 1);
    srv.kara_prefix = kara_dir;

    database_config_queue_default(srv.db);
    if (!env_get(LKT_ENV_RESTART) && autoclear)
        database_queue_clear(srv.db);
    lkt_queue_make_available(&srv.queue, lkt_event_prop);

    {   /* Module initialization */
        char *module = safe_malloc(sizeof(char) * PATH_MAX);
        if (!module) {
            LOG_ERROR("GENERAL", "Out of memory");
            return 1;
        }
        database_config_get_text(srv.db, "player", "module", module, PATH_MAX);
        reg_import(module, &srv.window_mod.reg, &srv.window_mod.handle);
        database_config_get_text(srv.db, "repo", "module", module, PATH_MAX);
        reg_import(module, &srv.repo_mod.reg, &srv.repo_mod.handle);
        free(module);

        if (MOD_CALL(srv.repo_mod, "new", &srv.queue, srv.db))
            return 100;
        if (MOD_CALL(srv.window_mod, "new", &srv.queue, srv.db))
            return 101;
    }

    /* Get ENV */
    /* Not working -> race condition with player module */
    char *env_current = env_get(LKT_ENV_CURRENT);
    if (env_current && env_current[0] && !STR_MATCH(env_current, "NULL")) {
        LOG_INFO("INIT", "Restart playback from %s", env_current);
        lkt_queue_send(&srv.queue, lkt_event_play_pos, (void *) (size_t) strtol(env_current, NULL, 0));
    }

    LOG_INFO("GENERAL", "Lektor was %s",
             env_get(LKT_ENV_RESTART) ? "restarted" : "started");

    lkt_listen(&srv);
    return EXIT_FAILURE;
}
